#app.py:

from BottleSessions import BottleSessions
from bottle import Bottle, request

app = Bottle()
btl = BottleSessions(app)

@app.route('/set/<key>/<val>')
def set_sess(key,val=None):

    request.session[key] = val
    return {key: val}

@app.route('/get/<key>')
def get_sess(key=None):

    return {key: request.session.get(key,'does not exist')}

@app.route('/')
def hello():
    return 'hello world'

if __name__ == '__main__':
    app.run()



import os
import sys
from secrets import token_urlsafe

from .backing import Backing
from .Session import Session

from bottle import request, response, PluginError

DEBUG = os.environ.get('DEBUG_SESSIONS', False)

default_backing_options = {
    'cache_type': 'SimpleCache',
}

def dprint(*args):
    """Debug Print"""
    if DEBUG:
        print(*args, file=sys.stderr)


class BottleSessions:
    """
    Create a BottleSession

    BottleSession(
            app, 
            session_backing, 
            session_cookie, 
            session_expire,
            session_secure,
            session_path,
            session_reference
            )

    Install middleware to provide sessions that persist accross http requests for a given
    client using http cookies and a backing object.

    Arguments:

    app:                the Bottle object - if provided instance self-installs 
                        default: None - instance does not self-install

    session_backing:    persistent backing for session objects
                        default: SimpleCache - creates a memory based cache
    
    session_cookie:     the name of the cookie used to track the session
                        default: 'bottlecookie'
                        environment: SESSION_COOKIE overrides parameter
    
    session_expire:     cookie life (and cache) life of the session object
                        default: 300 Sec (from last session update)
                        environment: SESSION_LIFE overrides parameter
    
    session_secure:     set the cookie with the 'Secure' flag
                        default: True (also true in bottle debug mode)
                        environment: SESSION_SECURE overrides parameter
    
    session_path:       http URL path the browser will provide the cookie
                        default '/' - any path on the site
                        environment: SESSION_PATH overrides parameter

    session_reference:  the name used to attach the session to the request object
                        default is 'session' (e.g. request.session )
                        environment: none

    If multiple BottleSessions are used in the same app, both session_reference and 
    session_cookie must be unique (this precludes using multiple sessions with environment 
    variables.)

    """
    name = 'BottleSession'      # Bottle Plugin API v2 required (this is updated)
    api = 2                     # Bottle Plugin API v2 required
    
    def __init__(self,
            app = None,                         # if app exists, module self-installs
            session_backing=None,               # Backing object, dict of config, or None
            session_cookie='bottlecookie',      # name of session cookie
            session_expire= 300,                # default ttl for cache and cookie
            session_secure= True,               # Secure Cookie
            session_path='/',                   # Cookie path
            session_reference='session'         # name of request reference to add for session
        ):
        """
        BottleSession.__init__()

        - establishes backing store
        - manage cookie specifications
        - installs self if app was provided
        """
        # This is the name of this plugin, and the name set for the
        # session when added to the request.
        self.name = session_reference

        if session_backing is None:
            self.backing = Backing(**default_backing_options)

        elif type(session_backing) is dict:
            self.backing = Backing(**session_backing)

        else:
            self.backing = session_backing
        
        self.cookie_name = os.environ.get('SESSION_COOKIE', session_cookie)
        self.cookie_expire = int(os.environ.get('SESSION_LIFE', session_expire))
        self.cookie_secure = os.environ.get('SECURE_COOKIE', session_secure)
        self.session_path = os.environ.get('SECURE_PATH', session_path)

        # self-install session middleware
        if app:
            app.install(self)

        print(f'*** Session "{self.name}" initialized using "{self.backing.cache_type}"')
        

    def setup(self, app):
        """
        setup()

        - Bottle plugin v2 api required
        - runs on plugin install
        - validate unique cookie names for multiple instances
        - validate unique session_reference names for multiple instances
        """

        for plugin in app.plugins:
        
            if isinstance(plugin, BottleSessions):
                # Multiple instantiations are permited with unique names and cookies.
                if plugin.name == self.name:
                    # name must be unique
                    emsg = f'\n\nError: BottleSession Duplicate name (session_reference) "{self.name}"\n'
                    raise PluginError(emsg)

                if plugin.cookie_name == self.cookie_name:
                    # cookies must be unique
                    emsg = f'\n\nError: BottleSession Duplicate cookie "{self.cookie_name}" allocated for "{plugin.name}" and "{self.name}"\n'
                    raise PluginError(emsg)
    

    def close(self):
        """
        close()

        Bottle plugin v2 api required
        - runs on plugin uninstall
        - nothing to do
        """
        pass


    def apply(self, f, context):
        """
        apply()

        - Bottle plugin v2 api required 
        - runs to decorate each applied path
        - adds wrappers to views
        - wrapper adds session_reference to request object
        """

        def session_wrapper(*args, **kwargs):
            """
            Middleware code in request path
                - open the session
                - call the view
                - save and close the session
            """

            session = self.open_session()
            
            # next middleware or run the view
            body =  f(*args, **kwargs)

            self.close_session(session)
            
            return body

        session_wrapper.__name__ = f.__name__
        return session_wrapper


    def set_session_cookie(self, session):
        """
        set_session_cookie()

        - runs on response path of route decoration
        - set cookie if it was updated to backing store
        - expire deleted cookies

        """

        if session.session_is_saved or session.session_deleted:
            # why bother reset the cookie life if we don't write clean 
            # sessions to the backing store?

            t = -1 if session.session_deleted else self.cookie_expire
            
            response.set_cookie(
                name=self.cookie_name,
                value=session.session_id,
                max_age = t,
                httponly = True,
                path = self.session_path,
                secure = self.cookie_secure
            )


    def open_session(self):
        """
        open_session()

            - retrieves session cookie from request
            - retrive session object from backing store
            - sets session_reference on request object
            - return session to caller
        """

        key = request.get_cookie(self.cookie_name)
        session = Session(backing=self.backing, session_id=key)
        
        # attach to request for use by views or successive middleware            
        setattr(request, self.name, session)
        return session
    

    def close_session(self,session):
        """
        close_session()

            - save changed sessions to backing store
            - updates client cookie in each response
        """
        # save the session
        session.session_save(expire=self.cookie_expire)

        # Add the Set-Cookie header        
        if session.session_is_saved or session.session_deleted:
            # why bother reset the cookie life if we don't write clean 
            # sessions to the backing store?

            t = -1 if session.session_deleted else self.cookie_expire
            
            response.set_cookie(
                name=self.cookie_name,
                value=session.session_id,
                max_age = t,
                httponly = True,
                path = self.session_path,
                secure = self.cookie_secure
            )



import os
import sys
from bottle import request, Bottle
from BottleSessions import BottleSessions

DEBUG=os.environ.get('DEBUG',False)

dprint = lambda *args, **kwargs: print(*args, **kwargs, file=sys.stderr)
app = Bottle()

CACHE_TYPE = os.environ.get('CACHE_TYPE', 'SimpleCache')

if CACHE_TYPE == 'SimpleCache':
    #
    # By default, cachelib SimpleCache is used for session backing.
    #  - this is non-persistent.
    #  - options are all defaults (default_timeout=300 threshold=500)
    #    but you can set these as either kwargs or session_backing
    #
    btl = BottleSessions(app)

elif CACHE_TYPE == 'FileSystem':
    #
    # The session backing can be explicitly established
    # This permits it to be put to other use by your application.
    #
    # the 'cache_type=' argument selects the cachelib type. All other arguments are passed to 
    # cachelib. Each cachelib type has some required and some optional configuration arguments.
    # 
    # Pass this object when creating the BottleSessions object using the 'session_backing='
    # argument.
    #
    from BottleSessions import Backing

    file_backing = Backing(cache_type='FileSystem', cache_dir='./.cache_dir')
    btl = BottleSessions(app, session_backing=file_backing)

elif CACHE_TYPE == 'Redis':
    #
    # You can pass all of the Backing options to BottleSession as a dict using the 'session_backing'
    # option, and BottleSessions will create the Backing object instance.
    #
    btl = BottleSessions(app, session_backing={
                'cache_type': 'Redis',
                'host': 'localhost',    # This is the default for cachelib redis
                'port' : '6379',        # This is the default for cachelib redis
            })

elif CACHE_TYPE == 'Memcached':
    #
    # If you want to explictly install BottleSessions
    # don't pass 'app', and install it with app.install()
    # - useful if you want your middleware installed in
    #   a particluar order
    #
    memcache_options = {
        'servers': ['localhost:11211'],
        'key_prefix': 'MyAppPrefix',
        'cache_type': 'Memcached'
    }

    btl = BottleSessions(session_backing=memcache_options)
    # do other things here...
    app.install(btl)

else:
    emsg = f'CACHE_TYPE "{CACHE_TYPE}" is unsupported'
    raise Exception(emsg)
#
# Multiple instantiations of BottleSession can be made, however the session_reference
# and session_cookie must be different
#

@app.route('/')
def index():
    """Report the current counter"""

    return f"The counter is at {request.session.get('cnt',0)}"


@app.route('/p')
def add_one():
    """Add one to the current session counter"""
    
    sess = request.session
    cnt = sess.get('cnt',0) +1
    sess['cnt'] = cnt

    return f'counter is up to {cnt}'    


@app.route('/n')
def sub_one():
    """Subtract one from the current session counter"""

    sess = request.session
    cnt = sess.get('cnt', 0) - 1
    sess['cnt'] = cnt

    return f'counter is down to {cnt}'


if __name__ == '__main__':
    app.run(host='localhost', port=8001, debug=DEBUG)


import json
import os
import sys
from secrets import token_urlsafe

from cachelib.simple import SimpleCache

from .backing import Backing


DEBUG = os.environ.get('DEBUG_SESSIONS', False)
TOKENLEN = 16      # Length of session key

def dprint(*args):
    """Debug Print"""
    if DEBUG:
        print(*args, file=sys.stderr)

def ddprint(*args):
    if DEBUG == '2':
        print(*args, file=sys.stderr)


default_backing = None
default_backing_config = {
    'cache_type': 'SimpleCache',
}

def _get_default_backing(backing):
    """
    _get_default_backing(backing)
    
    Returns the prefered backing store 
    - if user provides a valid Backing object, use it
    - if there is a default_backing object instantiated, use it
    - if the user provided a configuration dict, use it to create
      a new default_backing object
    - otherwise, create a default_backing object using our defaults.
    
    """
        # Probably they didn't mean to do this...
    global default_backing, default_backing_config

    if isinstance(backing, Backing):
        return backing

    if default_backing:
        return default_backing
    
    elif type(backing) is dict:
        default_backing = Backing(**backing)    

    else:
        # create a new default backing
        default_backing = Backing(**default_backing_config)
        
    return default_backing


class Session(dict):
    """
    
    session = Session(backing, sess_id)

    Each session is an object (dict) backed with persistent storage such that it 
    can be recalled on subsequent requests with its key.  Sessions are lazy loaded
    from backing.

    Arguments:

    session_backing:    The Backing object storing the desired key
                        default: None - creates a new, default backing store.

    session_id:         The session id is the key of the object to be retrieved.
                        default: None - creates a new key and object

    Sessions are saved to backing store as json if they have been modified, via
    session.save_session().  Only modified sessions are saved.  
    
    Updating a session dict (e.g. session['user'] = 'alice' or session.clear()) sets 
    them as modified. They can also be explicitly set as modified explicitly by setting 
    session.modified = True. 
    
    This may be required for indirect updates.
    e.g.
        session['user_attributes']['email'] = 'alice@example.org'
        session.session_modified = True
    
    """
    
    def __init__(self, backing=None, session_id=None):

        dict.__init__(self)

        self.session_backing = _get_default_backing(backing)
        self.session_id = session_id

        self.session_dict = False           # dict is loaded and valid
        self.session_modified =  False      # dict has been modified
        self.session_deleted =  False       # dict has been deleted

        self.session_is_saved = False       # useful for when to set cooikes

    #
    # session is loaded from backing store and marked valid only if it is accessed.
    # 
    # session is marked dirty if it modifed.
    #
    def __contains__(self, key):
        """
        session.__contains__()
        
        membership: x in session:
        - assure the dict is loaded
        - run the super call
        """
        ddprint(f'Session: contains {key}')

        if not self.session_dict: self.session_lazy_load()

        return super().__contains__(key)
    

    def __iter__(self):
        """
        session.__iter__()

        Iterator: For x in session:
        - assure the dict is loaded
        - run the super call
        """
        ddprint(f'Session: iter')

        if not self.session_dict: self.session_lazy_load()

        return super().__iter__()


    def __getitem__(self, key):
        """
        session.__getitem__()

        Get item from dict:
        - assure the dict is loaded
        - run the super call
        """
        ddprint(f'Session: getitem {key}')

        # lazy load may be required
        if not self.session_dict: self.session_lazy_load()
        
        return super().__getitem__(key)


    def __setitem__(self, key, val):
        """
        session.__setitem__()

        Set item in dict
        - assure the dict is valid
        - mark session as dirty
        - run the super call
        """
        ddprint(f'Session setitem {key} = {val}')
        
        # lazy load may be required
        if not self.session_dict: self.session_lazy_load()

        # Mark session dirty
        self.session_modified = True

        return super().__setitem__(key, val)


    def __getattribute__(self, name):
        """
        Handle dict method calls
        - handle 'session_*' calls with super
        - otherwise assure dict is valid
        - routines that write set modified flag
        - run the super
        """
        ddprint(f'Session: getattr {name}')

        if name.startswith('session_'):
            # escape geattr loops
            return super().__getattribute__(name)
        
        # lazy load may be required
        if not self.session_dict: self.session_lazy_load()

        # if a method is called that updates, set the session modified
        if name in ['clear', 'pop', 'popitem', 'setdefault', 'update']:
            self.session_modified = True

        return super().__getattribute__(name)

    
    def __str__(self):
        """
        session.__str__()

        Displayable string showing the session details
        """

        if self.session_dict:
            sdict = super().__str__() + ' >'
        else:
            sdict = '[ No data cached ]'

        tstat = f'<Session: [{self.session_id}] (' 
        tstat += ' dirty ' if self.session_modified else '' 
        tstat += ' deleted ) ' if self.session_deleted else ')'

        return tstat + sdict

    #
    # Internal methods to manage creating or reconstition of sessions from backing store
    #
    def session_lazy_load(self):
        """
        self.session_lazy_load()
        
        Load the session from backing store
        - verify we are not deleted
        - using either session_load() or session_new()

        """

        if self.session_deleted:
            raise Exception(f'Referencing deleted session {self.session_id}')

        # lazy load the session
        if self.session_id:
            dprint(f'Session: {self.session_id} attempting lazy load')
            self.session_load()

        else:
            dprint(f'Session: creating new key and session')
            self.session_new()    


    def session_load(self):
        """
        session.session_load()

        Load (or reload) session state from backing store
        - retrieves 'key' from backing store
        - deserializes the json into the session dict
        - set dict_valid if everything works
        - tempermental situations cause a session_new created
        """
        
        dprint(f'Session: {self.session_id} loading data')
        serial = self.session_backing.get(self.session_id)
        
        if serial:
            # just in case it fails to deserialize
            try:
                data_dict = json.loads(serial)

                super().clear()
                super().update(data_dict)

                self.session_dict = True
                self.session_modified = False

                dprint(f'Session: {self.session_id} load complete')

                return

            except:
                # data is bad - we'll pretend it's a new session
                dprint(f'Session: {self.session_id} failed to reload data')

        else:
            # Not found in backing store - create a new session
            dprint(f'Session: {self.session_id} no backing store')

        dprint(f'Session: {self.session_id} creating new session for this key')

        self.session_new()


    def session_new(self):
        """
        session.session_new()

        Generate a new session
        - Reuse an existing or cerate a sess_id
        - set session_dict
        - create an empty dict. 
        """

        if self.session_id is None:
            self.session_id = token_urlsafe(TOKENLEN)
        
        super().clear()

        self.session_dict = True

        dprint(f'Session: {self.session_id} new session created')

    #
    # Methods for Session users:  delete(), clear(), and save_session()
    #     
    def session_delete(self):
        """
        session.session_delete()

        Mark this session to be deleted
        - remove from backing store
        - set deleted flag to prevent flushing
        - mark as invalid
        """

        if self.session_dict is None:
            self.session_backing.delete(self.session_id)
        
        self.clear()
        self.session_dict = False
        self.session_deleted = True



    def session_save(self, expire=0, force=False):
        """
        session.session_save()

        Save session to backing store
        - only of not deleted set
        - must be modified or have force set
        - force does not flush a deleted session
        """
        
        if not self.session_deleted and (self.session_modified or force):
            # save this session to backing store
            serial = json.dumps(self)
            
            self.session_backing.set(self.session_id, serial, expire)
            self.session_modified = False
            self.session_is_saved = True

            dprint(f'Session {self.session_id} - flushed to backing store')
        else:
            dprint(f'Session: {self.session_id} - not dirty/notflushed')
    

    def session_regenerate(self):
        """
        session.session_regenerate()

        Generate a new sessionid
        - assure dict is valid
        - set session_id to new id
        - mark session dirty
        - remove old session_id from backing store
        """

        if not self.session_dict : self.session_lazy_load()

        old_session_id = self.session_id
        self.session_id = token_urlsafe(TOKENLEN)
        self.session_modified = True
        self.session_backing.delete(old_session_id)




"""
Backing

Backing offers abstraction of Pallets' cachelib, the underlying cache used, and
is to minimize changes required to utilize and change the cache selection without
needing to update code, just configuration data.

'SimpleCache'   ->  cachelib.SimpleCache.
                Default, non-persistant caching.

'FileSystem'    ->  cachelib.FileSystemCache.
                Persistant caching to a filesystem.

                The argument 'cache_dir' is required, others are optional.

'Redis'         ->  cachelib.RedisCache
                Persistant, potentially network based and shared.

                By default this uses a local instance of redis.

'Memcached'     ->  cachelib.MemcachedCache
                Persistant, potentially network based and shared.

                Requires 'servers' argument, e.g. servers=['localhost:11211']

Generally cachelib default expiration is 300 seconds.

Methods:

Underlying methods of cachelib.BaseCache:

    All cachelibs support at least delete(), get(), has(), and set().
    For most uses, these are what you need.
    
    Some cachelibs also support add(), clear(), dec(), inc(),
    delete_many(), get_many(), and set_many()

Refer to https://cachelib.readthedocs.io/ for details to see which methods are
implemented by each system and detailed explainations.  

"""
import sys
from cachelib import SimpleCache, RedisCache, MemcachedCache, FileSystemCache


_caches = {
    'FileSystem': FileSystemCache, 
    'Memcached': MemcachedCache,
    'Redis': RedisCache, 
    'SimpleCache': SimpleCache
    }

class Backing:
    
    def __init__(self, cache_type='SimpleCache', **kwargs):
        """
        back = Backing(cache_type, **kwargs)

        cache_type 
        - this can be 'FileSystem', 'Memcached', 'Redis', or (default) 'SimpleCache'
        - selects the associated Pallets cachelib and instantiates it
        
        kwargs
        - keyword arguments appropriate for the selected cache_type and are passed to cache_lib
        - Each type may have a different set of required and optional arguments based on the
          underlying technology.  See the cachelib documentation.

        """

        if cache_type not in _caches:
            raise Exception(f'Backing: cache_type "{cache_type}" unsupported')

        CacheBase = _caches[cache_type]
        self.cache = CacheBase(**kwargs)

        self.cache_type = cache_type

        print(f'INFO: {__name__} cache "{cache_type}" established', file=sys.stderr)

    
    def __getattr__(self, name):
        """Retrieve attr from underlying cache """

        return getattr(self.cache, name)
